#include <stdio.h>
#include <stdlib.h>

#define EPS 1E-10 //precision for the convergence of the carryover opinion
#define REP 1000 // max number of iterations for the convergence of the carryover opinion


/*
./carop graph.bin n1 n2 n3 n4...
Computes the carryover opinion of node n1 to all nodes and write the output in a file named n1.
Do the same for n2, n3 and n4...
*/

typedef struct {
	unsigned dim;	//dimensions of the squared matrix
	unsigned *degcum;	//cumulative degree
	unsigned *neigh;	//list of neighbors
	double *weight;	//Weights
} sparse;

sparse *file2sparse(char* file_name) {
	sparse *mat=malloc(sizeof(sparse));
	unsigned n,e;
	double t;
	FILE *file;
	
	printf("Charging graph from file %s\n",file_name);

	file=fopen(file_name,"r");
	fread(&(mat->dim),sizeof(unsigned),1,file);
	n=mat->dim;
	mat->degcum=malloc((n+1)*sizeof(unsigned));
	fread(mat->degcum,sizeof(unsigned),n+1,file);
	e=mat->degcum[n];
	mat->neigh=malloc(e*sizeof(unsigned));
	fread(mat->neigh,sizeof(unsigned),e,file);

	fread(&t,sizeof(double),1,file);
	if (feof(file)!=0) {
		mat->weight=NULL; 
		printf("Unweighted graph charged\n");
	}
	else {
		mat->weight=malloc(e*sizeof(double));
		mat->weight[0]=t;
		fread(mat->weight+1,sizeof(double),e-1,file);
		printf("Weighted graph charged\n");
	}
	fclose(file);
	return mat;
}

/*compute the vector matvect, product of the matrix mat and vector vect*/
void prodvect(sparse *mat,double *vect,double *matvect) {
	unsigned n,i,j;
	n=mat->dim;
	if (mat->weight==NULL) {
		for (i=0;i<n;i++) {
			matvect[i]=0;
			for (j=mat->degcum[i];j<mat->degcum[i+1];j++) {
				matvect[i]+=vect[mat->neigh[j]];
			}
		}
	}
	else {
		for (i=0;i<n;i++) {
			matvect[i]=0;
			for (j=mat->degcum[i];j<mat->degcum[i+1];j++) {
				matvect[i]+=vect[mat->neigh[j]]*mat->weight[j];
			}
		}
	}
}



int main(int argc,char** argv){
	sparse *mat;
	double *deg, *one, *x1, *x2, *x3;
	double minx, minx1, dx, maxdx;
	unsigned n, i, a, node;
	unsigned int r;
	FILE *file;

	mat=file2sparse(argv[1]);
	n=mat->dim;

	deg=malloc(n*sizeof(double));
	if (mat->weight==NULL) {
		for (i=0;i<n;i++) {
			deg[i]=mat->degcum[i+1]-mat->degcum[i];
		}
	}
	else {
		one=malloc(n*sizeof(double));
		for (i=0;i<n;i++) {
			one[i]=1;
		}
		prodvect(mat,one,deg);
		free(one);
	}

	for (i=0;i<n;i++) {
		deg[i]=(deg[i]==0)?1:deg[i];
	}


	x1=calloc(n,sizeof(double));
	x2=malloc(n*sizeof(double));

	for (a=2;a<argc;a++) {
		
		node=atoi(argv[a]);
		printf("Computing carryover opinion from node %u:\n",node);
		x1[node]=1.,r=0;
		do {

			maxdx=0., minx=1;
			
			/* step 1: averaging */
			prodvect(mat,x1,x2);
			for (i=0;i<n;i++) {
				x2[i]/=deg[i];
				minx=(minx>x2[i] && x2[i]>0) ? x2[i]:minx;
			}
			/* step 2: resetting */
			x2[node]=1.;

			/* step 3: rescaling */
			if (r>7){
				minx1=1-minx;
				for (i=0;i<n;i++) {
					if (x2[i]>0) {
						x2[i]=(x2[i]-minx)/minx1;
						dx=x2[i]-x1[i],	dx=(dx>-dx) ? dx : -dx, maxdx = (maxdx>dx) ? maxdx : dx;
					}
				}
			}
			else maxdx=EPS+1;

			x3=x1,x1=x2,x2=x3;
			if (++r%10==0) {
				printf("iteration: %u, error: %.2le\n",r,maxdx);
			}
		} while (maxdx>EPS && r<REP);

		printf("iteration: %u, error: %.2le\n",r,maxdx);

		printf("Writing the result for node %u\n",node);
		file=fopen(argv[a],"w");
		for (i=0;i<n;i++) {
			fprintf(file,"%u %le\n",i,x1[i]);
		}
		fclose(file);
		
	}
	return 0;
}


